#include <regex>

#include "zswagcl/private/openapi-config.hpp"
#include "zswagcl/export.hpp"
#include "stx/format.h"

namespace zswagcl
{

ZSWAGCL_EXPORT const std::string ZSERIO_OBJECT_CONTENT_TYPE = "application/x-zserio-object";
ZSWAGCL_EXPORT const std::string ZSERIO_REQUEST_PART = "x-zserio-request-part";
ZSWAGCL_EXPORT const std::string ZSERIO_REQUEST_PART_WHOLE = "*";

}